/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#ifdef _USRDLL
  #define _QUBFITS_PY_
#endif
#include "_qubfits.h"

#if defined(_WIN32)

#include <windows.h>
#include "notemplatewarning.h"
//#include <hash_map>
//using stdext::hash_map;

#else

#include <ext/hash_map>
using __gnu_cxx::hash_map;
namespace __gnu_cxx
{
        template<> struct hash< std::string >
        {
                size_t operator()( const std::string& x ) const
                {
                        return hash< const char* >()( x.c_str() );
                }
        };
        template<> struct hash< pair<int, std::string> >
        {
                size_t operator()( const pair<int, std::string>& x ) const
                {
                        return x.first/2 + hash<std::string>()( x.second )/2;
                }
	};
}

// linux .so loading
//      void* dlopen("path/to/lib.so", RTLD_NOW) [check NULL return]
//      char* dlerror();                         [get error str / clear any stale error]
//      void* dlsym(handle, "symbol")            [check NULL or dlerror()]
//            dlclose(handle)                    [ref-counted]

#include <dlfcn.h>

#endif

#include <map>

typedef pair<int, std::string> intstr;
struct intstrltpred
{
   bool operator()(const intstr& a, const intstr& b) const
   {
     if (a.first == b.first)
       return strcmp(a.second.c_str(), b.second.c_str()) < 0;
     else
       return a.first < b.first;
   }
};

PyMethodDef qubfits_module_methods[] = {
  // module-scope functions
  {"LoadPlugin", qubfits_LoadPlugin, METH_VARARGS,               "Load plugin from a .so file."},
  {"CreateCurve", qubfits_CreateCurve, METH_VARARGS,             "Create a new curve by name."},
  {"ListCurves", qubfits_ListCurves, METH_VARARGS,               "List the names of installed curves."},
  {"CreateFitter", qubfits_CreateFitter, METH_VARARGS,           "Create a new curve by name."},
  {"ListFitters", qubfits_ListFitters, METH_VARARGS,             "List the names of installed fitters."},
  {"UseData", qubfits_UseData, METH_VARARGS,                     "Transform data according to 'use' settings."},
  {"ChiSqrUncertainty", qubfits_ChiSqrUncertainty, METH_VARARGS, "Compute the error limits for a curve param."},
  {NULL, NULL, 0, NULL}
};

PyTypeObject qubfits_curve_type = {
  PyObject_HEAD_INIT(NULL)
  0,
  "_qubfits.Curve",
  sizeof(qubfits_curve),
  0,
  qubfits_curve_free,
  0,          /*tp_print*/
  qubfits_curve_getattr,
  qubfits_curve_setattr,
  qubfits_curve_compare,
  qubfits_curve_repr,
  0,          /*tp_as_number*/
  0,          /*tp_as_sequence*/
  0, // &qubfits_curve_asMapping,
  0,          /*tp_hash */
  0,          /*tp_call */
  qubfits_curve_str
};

PyMethodDef qubfits_curve_methods[] = {
  {"addEvent", qubfits_curve_addEvent, METH_VARARGS,             "Add a callback function for a named event."},
  {"delEvent", qubfits_curve_delEvent, METH_VARARGS,             "Remove a callback function."},
  {"setExtraVars", qubfits_curve_setExtraVars, METH_VARARGS,     "Assign names to extra data signals."},
  {"getParamName", qubfits_curve_getParamName, METH_VARARGS,     "Get the name of a param."},
  {"getParamVal", qubfits_curve_getParamVal, METH_VARARGS,       "Get the value of a param."},
  {"setParamVal", qubfits_curve_setParamVal, METH_VARARGS,       "Set the value of a param."},
  {"getParamConst", qubfits_curve_getParamConst, METH_VARARGS,   "Is the param held constant (not fitted)?"},
  {"setParamConst", qubfits_curve_setParamConst, METH_VARARGS,   "Set whether the param should be held constant."},
  {"initParams", qubfits_curve_initParams, METH_VARARGS,         "Set params to default or data-sensitive values (not implemented)."},
  {"evalAll", qubfits_curve_evalArrayAll, METH_VARARGS,          "Evaluate the curve (all components)."},
  {"evalOne", qubfits_curve_evalArrayOne, METH_VARARGS,          "Evaluate the curve (one component)."},
  {"evalDeriv", qubfits_curve_evalDerivArray, METH_VARARGS,      "Evaluate the curve's derivative wrt a param."},
  {"updateState", qubfits_curve_updateState, METH_VARARGS,       "Modify state according to a (partial) state tree."},
  {NULL, NULL, 0, NULL}
};


PyTypeObject qubfits_fitter_type = {
  PyObject_HEAD_INIT(NULL)
  0,
  "_qubfits.Fitter",
  sizeof(qubfits_fitter),
  0,
  qubfits_fitter_free,
  0,          /*tp_print*/
  qubfits_fitter_getattr,
  0, //qubfits_fitter_setattr,
  qubfits_fitter_compare,
  qubfits_fitter_repr,
  0,          /*tp_as_number*/
  0,          /*tp_as_sequence*/
  0, // &qubfits_fitter_asMapping,
  0,          /*tp_hash */
  0,          /*tp_call */
  qubfits_fitter_str
};

PyMethodDef qubfits_fitter_methods[] = {
  {"addEvent", qubfits_fitter_addEvent, METH_VARARGS,         "Add a callback function for a named event."},
  {"delEvent", qubfits_fitter_delEvent, METH_VARARGS,         "Remove a callback function."},
  {"getParamName", qubfits_fitter_getParamName, METH_VARARGS, "Get the name of a param."},
  {"getParamVal",  qubfits_fitter_getParamVal, METH_VARARGS,  "Get the value of a param."},
  {"setParamVal",  qubfits_fitter_setParamVal, METH_VARARGS,  "Set the value of a param."},
  {"fit",          qubfits_fitter_fit, METH_VARARGS,          "Fit curve params to data."},
  {"stop",         qubfits_fitter_stop, METH_VARARGS,         "Interrupt a running fit."},
  {"getParamErr",  qubfits_fitter_getParamErr, METH_VARARGS,  "Get the error limits for a curve param."},
  {"updateState",  qubfits_fitter_updateState, METH_VARARGS,  "Modify state according to a (partial) state tree."},
  {NULL, NULL, 0, NULL}
};

// ******************************************************************************************************************
// curve and fitter can be owned by py

typedef PyObject* (*RefWrap)(void *obj, bool owned);
PyObject* RefWrapCurve(void *obj, bool owned)
{
  return qubfits_curve_wrap((IQUB_Curve*) obj, owned);
}
PyObject* RefWrapFitter(void *obj, bool owned)
{
  return qubfits_fitter_wrap((IQUB_Fitter*) obj, owned);
}

#if defined(_WIN32)
typedef map<int, PyObject*> RefMap;
#else
typedef hash_map<int, PyObject*> RefMap;
#endif
RefMap qubfits_refmap;

PyObject* qubfits_FindRef(void *obj, RefWrap onWrap, bool owned=false)
{
  RefMap::iterator ri = qubfits_refmap.find((int)obj);
  if ( ri == qubfits_refmap.end() ) {
    qubfits_refmap[(int)obj] = onWrap(obj, owned);
    ri = qubfits_refmap.find((int)obj);
  } else {
    Py_INCREF( (*ri).second );
  }
  return (*ri).second;
}

void qubfits_ClearRef(void *obj)
{
  RefMap::iterator ri = qubfits_refmap.find((int)obj);
  if ( ri != qubfits_refmap.end() ) {
    qubfits_refmap.erase(ri);
  }
}


// ******************************************************************************************************************

QUBFITS_PY_API void init_qubfits()
{
  initqubfits(); // it's ok if it was already called; just re-registers curves and fitters
  qubfits_curve_type.ob_type = &PyType_Type;
  qubfits_fitter_type.ob_type = &PyType_Type;
  
  // PyObject *qubfits =
  Py_InitModule("_qubfits", qubfits_module_methods);
}

typedef int (*PlugInitFunc)();

PyObject* qubfits_LoadPlugin(PyObject* self, PyObject* args)
{
  int result = -666;
  char *path;
  if (!PyArg_ParseTuple(args, "s:LoadPlugin", &path))
    return NULL;
  
#if defined(_WIN32)
  HINSTANCE lib = LoadLibrary(path);
  if ( ! lib ) {
    PyErr_SetString(PyExc_Exception, "Can't load qubfits plugin");
    return NULL;
  }
  PlugInitFunc init = (PlugInitFunc) GetProcAddress(lib, "_initqubfits_plugin@0");
  if ( ! init ) {
    PyErr_SetString(PyExc_Exception, "Invalid qubfits plugin");
    return NULL;
  }
  try {
    result = init();
  }
  catch (...) {}
#else
//      void* dlopen("path/to/lib.so", RTLD_NOW) [check NULL return]
//      char* dlerror();                         [get error str / clear any stale error]
//      void* dlsym(handle, "symbol")            [check NULL or dlerror()]
//            dlclose(handle)                    [ref-counted]
  void *lib = dlopen(path, RTLD_NOW);
  if ( ! lib ) {
    PyErr_SetString(PyExc_Exception, dlerror());
    return NULL;
  }
  PlugInitFunc init = (PlugInitFunc) dlsym(lib, "initqubfits_plugin");
  if ( ! init ) {
    PyErr_SetString(PyExc_Exception, dlerror());
    return NULL;
  }
  try {
    result = init();
  }
  catch (...) {}
#endif
  
  return PyInt_FromLong(result);
}

PyObject* qubfits_CreateCurve(PyObject* self, PyObject* args)
{
  char *name = "";
  if (!PyArg_ParseTuple(args,"|s:CreateCurve", &name))
    return NULL;
  IQUB_Curve* result;
  Py_BEGIN_ALLOW_THREADS;
  result = CreateCurve(name);
  Py_END_ALLOW_THREADS;
  return qubfits_FindRef(result, RefWrapCurve, true);
}

PyObject* qubfits_ListCurves(PyObject* self, PyObject* args)
{
  QTR_Impl* lst;
  Py_BEGIN_ALLOW_THREADS;
  lst = ListCurves();
  Py_END_ALLOW_THREADS;
  
  PyObject *result = PyList_New(0);
  for (QUB_TreeMonoIter li = QUB_Tree(lst).children(); ! li->isNull(); ++li) {
    PyObject *pyName = PyString_FromString(li->dataAsString().c_str());
    PyList_Append(result, pyName);
    Py_DECREF(pyName);
  }

  QTR_DECREF(lst);
  return result;
}

PyObject* qubfits_CreateFitter(PyObject* self, PyObject* args)
{
  char *name = "";
  if (!PyArg_ParseTuple(args,"|s:CreateFitter", &name))
    return NULL;
  IQUB_Fitter* result;
  Py_BEGIN_ALLOW_THREADS;
  result = CreateFitter(name);
  Py_END_ALLOW_THREADS;
  return qubfits_FindRef(result, RefWrapFitter, true);
}

PyObject* qubfits_ListFitters(PyObject* self, PyObject* args)
{
  QTR_Impl* lst;
  Py_BEGIN_ALLOW_THREADS;
  lst = ListFitters();
  Py_END_ALLOW_THREADS;
  
  PyObject *result = PyList_New(0);
  for (QUB_TreeMonoIter li = QUB_Tree(lst).children(); ! li->isNull(); ++li) {
    PyObject *pyName = PyString_FromString(li->dataAsString().c_str());
    PyList_Append(result, pyName);
    Py_DECREF(pyName);
  }
  
  QTR_DECREF(lst);
  return result;
}

PyObject* qubfits_UseData(PyObject* self, PyObject* args)
{
  PyObject *pyData, *pyUse;
  if (!PyArg_ParseTuple(args, "OO:UseData(data, properties)", &pyData, &pyUse))
    return NULL;
  
  if ( ! (qubtree_node_check(pyData) && qubtree_node_check(pyUse)) ) {
    PyErr_SetString(PyExc_TypeError, "UseData: qubtree.Nodes expected");
    return NULL;
  }
  
  QUB_Tree treeData = qubtree_node_obj(pyData);
  QUB_Tree use = qubtree_node_obj(pyUse);
  QUB_Tree result;
  
  Py_BEGIN_ALLOW_THREADS;
  result = QF_UseData(treeData, use);
  Py_END_ALLOW_THREADS;
  
  return qubtree_node_using(result.getImpl());
}

PyObject* qubfits_ChiSqrUncertainty(PyObject* self, PyObject* args)
{
  PyObject *pyCurve, *pyData;
  int ipar;
  double deltaPct = 0.01;
  if (!PyArg_ParseTuple(args,"OOi|d:ChiSqrUncertainty", &pyCurve, &pyData, &ipar, &deltaPct))
    return NULL;
  
  if (!qubfits_curve_check(pyCurve)) {
    PyErr_SetString(PyExc_TypeError, "ChiSqrUncertainty arg 1: qubfits.Curve expected");
    return NULL;
  }
  IQUB_Curve *curve = qubfits_curve_obj(pyCurve);
  
  if (!qubtree_node_check(pyData)) {
    PyErr_SetString(PyExc_TypeError, "ChiSqrUncertainty arg 2: qubtree.Node expected");
    return NULL;
  }
  QUB_Tree treeData = qubtree_node_obj(pyData);

  double result;
  Py_BEGIN_ALLOW_THREADS
  QF_Data data;
  double *ff;
  TreeToQFData(treeData, &data, &ff);

  result = ChiSqrUncertainty(curve, &data, ff, ipar, deltaPct);

  CleanupQFData(&data);
  Py_END_ALLOW_THREADS;

  return PyFloat_FromDouble(result);
}


// ******************************************************************************************************************

class qubfits_curve_eventAdapter : public IQUB_EvtListener
{
public:
  PyObject *func;
  qubfits_curve_eventAdapter(PyObject* _func)
    : func(_func)
  {
    Py_INCREF(func);
  }
#if defined(_WIN32)
  virtual HRESULT __stdcall QueryInterface(const IID& iid, void ** ppvObject) { return NULL; }
  virtual ULONG __stdcall AddRef() { return -1; }
  virtual ULONG __stdcall Release() { return -1; }
#endif
  virtual void __cdecl QUBEvt(const char *name, void *sender)
  {
    PyGILState_STATE gstate = PyGILState_Ensure();
    PyObject* pysender = qubfits_FindRef(sender, RefWrapCurve);
    PyObject* args = Py_BuildValue("sO", name, pysender);
    PyObject* result = PyObject_Call(func, args, NULL);
    if ( ! result ) {
      PyErr_Clear();
    } else {
      Py_DECREF(result);
    }
    Py_DECREF(args);
    Py_DECREF(pysender);
    PyGILState_Release(gstate);
  }
  virtual ~qubfits_curve_eventAdapter()
  {
    Py_DECREF(func);
  }
};

#if defined(_WIN32)
typedef map< pair<int, string>, qubfits_curve_eventAdapter*, intstrltpred> CurveEventMap;
#else
typedef hash_map< pair<int, string>, qubfits_curve_eventAdapter*> CurveEventMap;
#endif

// ******************************************************************************************************************

QUBFITS_PY_API PyObject*    qubfits_curve_wrap(IQUB_Curve* curve, bool owned)
{
  qubfits_curve *pyCurve = PyObject_New(qubfits_curve, &qubfits_curve_type);
  pyCurve->impl = curve;
  pyCurve->owned = owned;
  pyCurve->events = new CurveEventMap;
  return (PyObject *) pyCurve;
}
  
QUBFITS_PY_API int          qubfits_curve_check(PyObject* pyCurve)
{
  return ( pyCurve->ob_type == &qubfits_curve_type );
}

QUBFITS_PY_API void         qubfits_curve_free(PyObject* pyObj)
{
  qubfits_curve *pyCurve = (qubfits_curve*) pyObj;
  qubfits_ClearRef(pyCurve->impl);
  if ( pyCurve->owned )
    pyCurve->impl->Free();
  CurveEventMap& events = * (CurveEventMap*) (pyCurve->events);
  for (CurveEventMap::iterator ei=events.begin(); ei!=events.end(); ++ei) {
    delete ei->second;
  }
  delete &events;
  PyObject_Del(pyCurve);
}

QUBFITS_PY_API IQUB_Curve*  qubfits_curve_obj(PyObject* pyCurve)
{
  return ( ((qubfits_curve*) pyCurve)->impl );
}

PyObject* qubfits_curve_repr(PyObject* self)
{
  return PyString_FromString(qubfits_curve_obj(self)->getName());
}

PyObject* qubfits_curve_str(PyObject* self)
{
  return qubfits_curve_repr(self);
}

int       qubfits_curve_compare(PyObject* v, PyObject* w)
{
  if ( qubfits_curve_check( w ) ) {
    long l1 = (long) qubfits_curve_obj( v );
    long l2 = (long) qubfits_curve_obj( w );
    if ( l1 < l2 )
      return -1;
    else if ( l1 == l2 )
      return 0;
    else
      return 1;
  }
  else
    return -1;
}
  
PyObject* qubfits_curve_addEvent(PyObject* self, PyObject* args)
{
  char *name;
  PyObject* func;
  if (!PyArg_ParseTuple(args, "sO:addEvent(name, func(name, curve))", &name, &func))
    return NULL;
  
  CurveEventMap& events = * (CurveEventMap*) ((qubfits_curve*) self)->events;
  qubfits_curve_eventAdapter *adapter = new qubfits_curve_eventAdapter(func);
  events[pair<int,string>((int)func, name)] = adapter;
  qubfits_curve_obj(self)->Events()->add(name, adapter);

  return Py_BuildValue("");
}

PyObject* qubfits_curve_delEvent(PyObject* self, PyObject* args)
{
  char *name;
  PyObject* func;
  if (!PyArg_ParseTuple(args, "sO:delEvent(name, func(name, curve))", &name, &func))
    return NULL;
  
  CurveEventMap& events = * (CurveEventMap*) ((qubfits_curve*) self)->events;
  CurveEventMap::iterator ei = events.find(pair<int,string>((int)func, name));
  if ( ei != events.end() ) {
    qubfits_curve_obj(self)->Events()->del(name, ei->second);
    delete ei->second;
    events.erase(ei);
  }

  return Py_BuildValue("");
}

// ******************************************************************************************************************

PyObject* qubfits_curve_getattr(PyObject* self, char* name)
  // r/w: eqn, componentCount
  // r:   name, canSetEqn, paramCount, hasDeriv, msg, state
{
  IQUB_Curve* curve = qubfits_curve_obj(self);
  string sName(name);

  // shall we skate on thin ice and assume none of these get the GIL?
  
  if ( sName == "eqn" )
    return PyString_FromString(curve->getEqn());
  else if ( sName == "componentCount" )
    return PyInt_FromLong(curve->getComponentCount());
  else if ( sName == "name" )
    return PyString_FromString(curve->getName());
  else if ( sName == "canSetEqn" )
    return PyBool_FromLong( curve->canSetEqn() );
  else if ( sName == "paramCount" )
    return PyInt_FromLong( curve->getParamCount() );
  else if ( sName == "hasDeriv" )
    return PyBool_FromLong( curve->hasDeriv() );
  else if ( sName == "msg" )
    return PyString_FromString( curve->getMessage() );
  else if ( sName == "state" ) {
    QTR_Impl* state = curve->getState();
    QUB_Tree tstate(state);
    QTR_DECREF(state);
    return qubtree_node_using( state );
  }
  else
    return Py_FindMethod(qubfits_curve_methods, self, name);
}

int       qubfits_curve_setattr(PyObject* self, char* name, PyObject* val)
{
  if (0 == strcmp(name, "eqn")) {
    if ( PyString_Check(val) )
      qubfits_curve_obj(self)->setEqn( PyString_AsString(val) );
    else {
  Py_BEGIN_ALLOW_THREADS
      PyErr_SetString(PyExc_TypeError, "Curve.eqn: expected a string");
  Py_END_ALLOW_THREADS
      return -1;
    }
  } else if (0 == strcmp(name, "componentCount")) {
    if ( PyInt_Check(val) )
  Py_BEGIN_ALLOW_THREADS
      qubfits_curve_obj(self)->setComponentCount(PyInt_AsLong(val));
  Py_END_ALLOW_THREADS
    else {
      PyErr_SetString(PyExc_TypeError, "Curve.componentCount: expected an int");
      return -1;
    }
  } else {
    PyErr_SetString(PyExc_NameError, "unknown or read-only property");
    return -1;
  }
  
  return 0;
}
      

// ******************************************************************************************************************

PyObject* qubfits_curve_setExtraVars(PyObject* self, PyObject* args)
{
  PyObject *pyData;
  if (!PyArg_ParseTuple(args,"O:qubfits.Curve.setExtraVars", &pyData))
    return NULL;
  
  if (!qubtree_node_check(pyData)) {
    PyErr_SetString(PyExc_TypeError, "setExtraVars: arg 1: qubtree.Node expected");
    return NULL;
  }
  QUB_Tree treeData = qubtree_node_obj(pyData);

  Py_BEGIN_ALLOW_THREADS
  QF_Data data;
  double *ff;
  TreeToQFData(treeData, &data, &ff);
  
  qubfits_curve_obj(self)->setExtraVars(data.vname, data.nv);

  CleanupQFData(&data);
  Py_END_ALLOW_THREADS;

  return Py_BuildValue("");
}

PyObject* qubfits_curve_getParamName(PyObject* self, PyObject* args)
{
  int ipar;
  if (!PyArg_ParseTuple(args, "i:qubfits.Curve.getParamName", &ipar))
    return NULL;
  
  return PyString_FromString( qubfits_curve_obj(self)->getParamName(ipar) );
}

PyObject* qubfits_curve_getParamVal(PyObject* self, PyObject* args)
{
  int ipar;
  if (!PyArg_ParseTuple(args, "i:qubfits.Curve.getParamVal", &ipar))
    return NULL;
  
  return PyFloat_FromDouble( qubfits_curve_obj(self)->getParamVal(ipar) );
}

PyObject* qubfits_curve_setParamVal(PyObject* self, PyObject* args)
{
  int ipar;
  double val;
  PyObject* tfo = NULL;
  if (!PyArg_ParseTuple(args, "id|O:qubfits.curve.setParamVal", &ipar, &val, &tfo))
    return NULL;
  
  int tf = (tfo == NULL) || PyObject_IsTrue(tfo);
  Py_BEGIN_ALLOW_THREADS;
  qubfits_curve_obj(self)->setParamVal(ipar, val, tf);
  Py_END_ALLOW_THREADS;
  
  return Py_BuildValue("");
}

PyObject* qubfits_curve_getParamConst(PyObject* self, PyObject* args)
{
  int ipar;
  if (!PyArg_ParseTuple(args, "i:qubfits.Curve.getParamConst", &ipar))
    return NULL;
  
  return PyBool_FromLong( qubfits_curve_obj(self)->getParamConst(ipar) );
}

PyObject* qubfits_curve_setParamConst(PyObject* self, PyObject* args)
{
  int ipar;
  PyObject* tfo;
  int tf;
  if (!PyArg_ParseTuple(args, "iO:qubfits.Curve.setParamConst", &ipar, &tfo))
    return NULL;
  tf = PyObject_IsTrue(tfo);
  
  Py_BEGIN_ALLOW_THREADS;
  qubfits_curve_obj(self)->setParamConst(ipar, tf);
  Py_END_ALLOW_THREADS;
  
  return Py_BuildValue("");
}
								     
PyObject* qubfits_curve_initParams(PyObject* self, PyObject* args)
{
  PyObject *pyData;
  if (!PyArg_ParseTuple(args,"O:qubfits.Curve.initParams", &pyData))
    return NULL;
  
  if (!qubtree_node_check(pyData)) {
    PyErr_SetString(PyExc_TypeError, "initParams: arg 1: qubtree.Node expected");
    return NULL;
  }
  QUB_Tree treeData = qubtree_node_obj(pyData);

  Py_BEGIN_ALLOW_THREADS
  QF_Data data;
  double *ff;
  TreeToQFData(treeData, &data, &ff);
  
  qubfits_curve_obj(self)->initParams(data.xx, data.yy, data.n);

  CleanupQFData(&data);
  Py_END_ALLOW_THREADS;

  return Py_BuildValue("");
}

PyObject* qubfits_curve_evalArrayAll(PyObject* self, PyObject* args)
{
  PyObject *pyData;
  if (!PyArg_ParseTuple(args,"O:qubfits.Curve.evalAll", &pyData))
    return NULL;
  
  if (!qubtree_node_check(pyData)) {
    PyErr_SetString(PyExc_TypeError, "evalAll: arg 1: qubtree.Node expected");
    return NULL;
  }
  QUB_Tree treeData = qubtree_node_obj(pyData);

  Py_BEGIN_ALLOW_THREADS
  QF_Data data;
  double *ff;
  TreeToQFData(treeData, &data, &ff);
  
  qubfits_curve_obj(self)->evalArrayAll(&data, ff);

  CleanupQFData(&data);
  Py_END_ALLOW_THREADS;

  return Py_BuildValue("");
}

PyObject* qubfits_curve_evalArrayOne(PyObject* self, PyObject* args)
{
  PyObject *pyData;
  int ic;
  if (!PyArg_ParseTuple(args,"Oi:qubfits.Curve.evalOne", &pyData, &ic))
    return NULL;
  
  if (!qubtree_node_check(pyData)) {
    PyErr_SetString(PyExc_TypeError, "evalOne: arg 1: qubtree.Node expected");
    return NULL;
  }
  QUB_Tree treeData = qubtree_node_obj(pyData);

  Py_BEGIN_ALLOW_THREADS
  QF_Data data;
  double *ff;
  TreeToQFData(treeData, &data, &ff);
  
  qubfits_curve_obj(self)->evalArrayOne(&data, ff, ic);

  CleanupQFData(&data);
  Py_END_ALLOW_THREADS;

  return Py_BuildValue("");
}

PyObject* qubfits_curve_evalDerivArray(PyObject* self, PyObject* args)
{
  PyObject *pyData;
  if (!PyArg_ParseTuple(args,"O:qubfits.Curve.evalDeriv", &pyData))
    return NULL;
  
  if (!qubtree_node_check(pyData)) {
    PyErr_SetString(PyExc_TypeError, "evalDeriv: arg 1: qubtree.Node expected");
    return NULL;
  }
  QUB_Tree treeData = qubtree_node_obj(pyData);

  Py_BEGIN_ALLOW_THREADS
  QF_Data data;
  double *ff;
  TreeToQFData(treeData, &data, &ff);
  
  int npar = qubfits_curve_obj(self)->getParamCount();
  QUB_Tree dparNode = treeData["dPar"];
  dparNode.setNumData(QTR_TYPE_DOUBLE, npar, 1, (void*)0);
  double *dpar = (double*) dparNode.data();
  for (int i=0; i<npar; ++i)
    dpar[i] = 0.0;
  
  qubfits_curve_obj(self)->evalDerivArray(&data, dpar);

  CleanupQFData(&data);
  Py_END_ALLOW_THREADS;

  return Py_BuildValue("");
}

PyObject* qubfits_curve_updateState(PyObject* self, PyObject* args)
{
  PyObject *pyState;
  if (!PyArg_ParseTuple(args, "O:qubfits.Curve.updateState", &pyState))
    return NULL;
  
  if (!qubtree_node_check(pyState)) {
    PyErr_SetString(PyExc_TypeError, "updateState: arg 1: qubtree.Node expected");
    return NULL;
  }
  QUB_Tree state = qubtree_node_obj(pyState);
  
  Py_BEGIN_ALLOW_THREADS;
  qubfits_curve_obj(self)->updateState(state.getImpl());
  Py_END_ALLOW_THREADS;
  
  return Py_BuildValue("");
}

// ******************************************************************************************************************

class qubfits_fitter_eventAdapter : public IQUB_EvtListener
{
public:
  PyObject *func;
  qubfits_fitter_eventAdapter(PyObject* _func)
    : func(_func)
  {
    Py_INCREF(func);
  }
  virtual ~qubfits_fitter_eventAdapter()
  {
    Py_DECREF(func);
  }
#if defined(_WIN32)
  virtual HRESULT __stdcall QueryInterface(const IID& iid, void ** ppvObject) { return NULL; }
  virtual ULONG __stdcall AddRef() { return -1; }
  virtual ULONG __stdcall Release() { return -1; }
#endif
  virtual void __cdecl QUBEvt(const char *name, void *sender)
  {
    PyGILState_STATE gstate = PyGILState_Ensure();
    PyObject* pysender = qubfits_FindRef(sender, RefWrapFitter);
    PyObject* args = Py_BuildValue("sO", name, pysender);
    PyObject* result = PyObject_Call(func, args, NULL);
    if ( ! result ) {
      PyErr_Clear();
    } else {
      Py_DECREF(result);
    }
    Py_DECREF(args);
    Py_DECREF(pysender);
    PyGILState_Release(gstate);
  }
};

#if defined (_WIN32)
typedef map< pair<int,string>, qubfits_fitter_eventAdapter*> FitterEventMap;
#else
typedef hash_map< pair<int,string>, qubfits_fitter_eventAdapter*> FitterEventMap;
#endif

// ******************************************************************************************************************

QUBFITS_PY_API PyObject*    qubfits_fitter_wrap(IQUB_Fitter* fitter, bool owned)
{
  qubfits_fitter *pyFitter = PyObject_New(qubfits_fitter, &qubfits_fitter_type);
  pyFitter->impl = fitter;
  pyFitter->owned = owned;
  pyFitter->events = new FitterEventMap;
  return (PyObject *) pyFitter;
}
  
QUBFITS_PY_API int          qubfits_fitter_check(PyObject* pyFitter)
{
  return ( pyFitter->ob_type == &qubfits_fitter_type );
}

QUBFITS_PY_API void         qubfits_fitter_free(PyObject* pyObj)
{
  qubfits_fitter* pyFitter = (qubfits_fitter*) pyObj;
  qubfits_ClearRef(pyFitter->impl);
  if ( pyFitter->owned )
    pyFitter->impl->Free();
  FitterEventMap& events = * (FitterEventMap*) pyFitter->events;
  for (FitterEventMap::iterator ei=events.begin(); ei!=events.end(); ++ei)
    delete ei->second;
  delete (FitterEventMap*) pyFitter->events;
  PyObject_Del(pyFitter);
}

QUBFITS_PY_API IQUB_Fitter* qubfits_fitter_obj(PyObject* pyFitter)
{
  return ( ((qubfits_fitter*) pyFitter)->impl );
}

PyObject* qubfits_fitter_repr(PyObject* self)
{
  return PyString_FromString(qubfits_fitter_obj(self)->getName());
}

PyObject* qubfits_fitter_str(PyObject* self)
{
  return qubfits_fitter_repr(self);
}

int       qubfits_fitter_compare(PyObject* v, PyObject* w)
{
  if ( qubfits_fitter_check( w ) ) {
    long l1 = (long) qubfits_fitter_obj( v );
    long l2 = (long) qubfits_fitter_obj( w );
    if ( l1 < l2 )
      return -1;
    else if ( l1 == l2 )
      return 0;
    else
      return 1;
  }
  else
    return -1;
}
  
PyObject* qubfits_fitter_addEvent(PyObject* self, PyObject* args)
{
  char *name;
  PyObject* func;
  if (!PyArg_ParseTuple(args, "sO:addEvent(name, func(name, fitter))", &name, &func))
    return NULL;
  
  FitterEventMap& events = * (FitterEventMap*) ((qubfits_fitter*) self)->events;
  qubfits_fitter_eventAdapter *adapter = new qubfits_fitter_eventAdapter(func);
  events[pair<int,string>((int)func, name)] = adapter;
  qubfits_fitter_obj(self)->Events()->add(name, adapter);

  return Py_BuildValue("");
}

PyObject* qubfits_fitter_delEvent(PyObject* self, PyObject* args)
{
  char *name;
  PyObject* func;
  if (!PyArg_ParseTuple(args, "sO:delEvent(name, func(name, fitter))", &name, &func))
    return NULL;
  
  FitterEventMap& events = * (FitterEventMap*) ((qubfits_fitter*) self)->events;
  FitterEventMap::iterator ei = events.find(pair<int,string>((int)func, name));
  if ( ei != events.end() ) {
    qubfits_fitter_obj(self)->Events()->del(name, ei->second);
    delete ei->second;
    events.erase(ei);
  }
  
  return Py_BuildValue("");
}

// ******************************************************************************************************************

PyObject* qubfits_fitter_getattr(PyObject* self, char* name)
  // r: name, paramCount, iterations, curve, msg, state
{
  IQUB_Fitter* fitter = qubfits_fitter_obj(self);
  string sName(name);
  
  // shall we skate on thin ice and assume none of these get the GIL?
  
  if ( sName == "name" )
    return PyString_FromString( fitter->getName() );
  else if ( sName == "paramCount" )
    return PyInt_FromLong( fitter->getParamCount() );
  else if ( sName == "iterations" )
    return PyInt_FromLong( fitter->getIterations() );
  else if ( sName == "curve" )
    return qubfits_FindRef( fitter->getCurve(), RefWrapCurve );
  else if ( sName == "msg" )
    return PyString_FromString( fitter->getMessage() );
  else if ( sName == "state" ) {
    QTR_Impl* state = fitter->getState();
    QUB_Tree tstate(state);
    QTR_DECREF(state);
    return qubtree_node_using( state );
  }
  else
    return Py_FindMethod(qubfits_fitter_methods, self, name);
}

// ******************************************************************************************************************

PyObject* qubfits_fitter_getParamName(PyObject* self, PyObject* args)
{
  int ipar;
  if (!PyArg_ParseTuple(args, "i:qubfits.Fitter.getParamName", &ipar))
    return NULL;
  
  return PyString_FromString( qubfits_fitter_obj(self)->getParamName(ipar) );
}

PyObject* qubfits_fitter_getParamVal(PyObject* self, PyObject* args)
{
  int ipar;
  if (!PyArg_ParseTuple(args, "i:qubfits.Fitter.getParamVal", &ipar))
    return NULL;
  
  return PyFloat_FromDouble( qubfits_fitter_obj(self)->getParamVal(ipar) );
}

PyObject* qubfits_fitter_setParamVal(PyObject* self, PyObject* args)
{
  int ipar;
  double val;
  if (!PyArg_ParseTuple(args, "id:qubfits.Fitter.setParamVal", &ipar, &val))
    return NULL;
  
  Py_BEGIN_ALLOW_THREADS;
  qubfits_fitter_obj(self)->setParamVal(ipar, val);
  Py_END_ALLOW_THREADS;
  
  return Py_BuildValue("");
}

PyObject* qubfits_fitter_fit(PyObject* self, PyObject* args)
{
  PyObject *pyCurve, *pyData;
  if (!PyArg_ParseTuple(args,"OO:Fitter.fit", &pyCurve, &pyData))
    return NULL;
  
  if (!qubfits_curve_check(pyCurve)) {
    PyErr_SetString(PyExc_TypeError, "Fitter.fit: arg 1: qubfits.Curve expected");
    return NULL;
  }
  IQUB_Curve *curve = qubfits_curve_obj(pyCurve);

  vector<QUB_Tree> treeData;
  if (qubtree_node_check(pyData)) {
    treeData.push_back( qubtree_node_obj(pyData) );
  } else if (PySequence_Check(pyData)) {
    int ndata = PySequence_Size(pyData);
    if (ndata == 0) goto BAD_DATA;
    for (int i=0; i<ndata; ++i) {
      PyObject* item = PySequence_GetItem(pyData, i);
      if (!qubtree_node_check(item)) {
	Py_DECREF(item);
	goto BAD_DATA;
      }
      treeData.push_back( qubtree_node_obj(item) );
      Py_DECREF(item);
    }
  } else {
  BAD_DATA:
    PyErr_SetString(PyExc_TypeError, "Fitter.fit: arg 2: qubtree.Node(s) expected");
    return NULL;
  }

  int result;
  Py_BEGIN_ALLOW_THREADS;
  vector<QF_Data> data(treeData.size());
  vector<QF_Data*> pdata(treeData.size());
  vector<double*> ff(treeData.size());
  for (int i=0; i<treeData.size(); ++i) {
    TreeToQFData(treeData[i], &(data[i]), &(ff[i]));
    pdata[i] = &(data[i]);
  }
  result = qubfits_fitter_obj(self)->Fit(curve, &(pdata[0]), pdata.size());

  for (int i=0; i<data.size(); ++i)
    CleanupQFData(&(data[i]));
  Py_END_ALLOW_THREADS;

  return PyInt_FromLong(result);
}

PyObject* qubfits_fitter_stop(PyObject* self, PyObject* args)
{
  Py_BEGIN_ALLOW_THREADS;
  qubfits_fitter_obj(self)->Stop();
  Py_END_ALLOW_THREADS;
  return Py_BuildValue("");
}

PyObject* qubfits_fitter_getParamErr(PyObject* self, PyObject* args)
{
  int ipar;
  if (!PyArg_ParseTuple(args, "i:Fitter.getParamErr", &ipar))
    return NULL;
  
  double result;
  Py_BEGIN_ALLOW_THREADS;
  result = qubfits_fitter_obj(self)->getParamErr(ipar);
  Py_END_ALLOW_THREADS;
  return PyFloat_FromDouble(result);
}

PyObject* qubfits_fitter_updateState(PyObject* self, PyObject* args)
{
  PyObject *pyState;
  if (!PyArg_ParseTuple(args, "O:qubfits.Fitter.updateState", &pyState))
    return NULL;
  
  if (!qubtree_node_check(pyState)) {
    PyErr_SetString(PyExc_TypeError, "updateState: arg 1: qubtree.Node expected");
    return NULL;
  }
  QUB_Tree state = qubtree_node_obj(pyState);
  
  Py_BEGIN_ALLOW_THREADS;
  qubfits_fitter_obj(self)->updateState(state.getImpl());
  Py_END_ALLOW_THREADS;
  
  return Py_BuildValue("");
}


// ******************************************************************************************************************
